class AuthNotFound(Exception):
    pass

class SessionExpired(Exception):
    pass

class SignError(Exception):
    pass

class UnknownError(Exception):
    pass

class CPUlimit(Exception):
    pass

class ExpiredTransaction(Exception):
    pass

class NotImplementedError(Exception):
    pass

class CookiesExpired(Exception):
    pass

class PayWithPushError(Exception):
    pass


import cloudscraper
from .contract import Contract
from .exceptions import PayWithPushError
from eospy.utils import sig_digest
from .types import Payers
class AtomicHub:
    """
    Allowed actions:
    - atomicassets
    - atomicmarket

    """
    def __init__(self, trx, network="mainnet"):
        self.trx = trx
        self.wax = trx.client.wax

        if self.trx.actions[0].result["account"] != "res.pink" or\
           self.trx.actions[0].result["name"] != "noop" or\
           self.trx.actions[0].result["authorization"][0]["actor"] != "res.pink" or\
           self.trx.actions[0].result["authorization"][0]["permission"] != "paybw":
            
            self.trx.actions = [
                Contract("res.pink", actor="res.pink", permission="paybw").noop()
            ] + self.trx.actions

        self.scraper = cloudscraper.create_scraper(browser={'custom': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/106.0.0.0 Safari/537.36'})

        self.sign_link = "https://wax-mainnet-signer.api.atomichub.io/v1/sign"
        self.push_link = "chain.push_transaction"

    def push(self) -> dict:
        signed = self.trx.get_trx_extend_info()
        signatures = signed['signatures']

        # sign with atomichub
        self.scraper.options(self.sign_link, headers={
            "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.9",
            "Accept-Encoding": "gzip, deflate",
            "Accept-Language": "ru-RU,ru;q=0.9,en-US;q=0.8,en;q=0.7",
            "Cache-Control": "max-age=0",
            "Connection": "keep-alive",
            "DNT": "1",
            "Host": "wax.api.atomicassets.io",
            "Upgrade-Insecure-Requests": "1",
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/90.0.4430.85 Safari/537.36"
        })

        sign_packed = self.scraper.post(self.sign_link, json={"transaction": signed['packed']}).json()

        if sign_packed.get('success') is False:
            raise PayWithPushError(sign_packed.get('message'))

        signatures += sign_packed['data']

        # push transaction
        push = self.wax.post(self.push_link, json={
            "signatures": signatures,
            "compression": 0,
            "packed_context_free_data": "",
            "packed_trx": signed['packed']
        })
        return push


class Nefty:
    def __init__(self, trx, network="mainnet"):
        self.trx = trx
        self.wax = self.trx.wax

        if self.trx.actions[0].result["account"] != "neftyblocksd" or\
           self.trx.actions[0].result["name"] != "paycpu" or\
           self.trx.actions[0].result["authorization"][0]["actor"] != "neftybrespay" or\
           self.trx.actions[0].result["authorization"][0]["permission"] != "active":
            
            self.trx.actions = [
                Contract("neftybrespay", actor="neftybrespay").paycpu()
            ] + self.trx.actions

        self.scraper = cloudscraper.create_scraper()
        self.scraper.headers.update({
            'accept': '*/*',
            'accept-encoding': 'gzip, deflate, br',
            'accept-language': 'ru-RU,ru;q=0.9,en-US;q=0.8,en;q=0.7',
            'access-control-request-headers': 'content-type',
            'access-control-request-method': 'POST',
            'cache-control': 'no-cache',
            'pragma': 'no-cache',
            'sec-fetch-dest': 'empty',
            'sec-fetch-mode': 'cors',
            'sec-fetch-site': 'same-site'
        })

        if network == "testnet":
            self.sign_link = "https://cpu-test.neftyblocks.com/"
            self.push_link = "https://wax-testnet.neftyblocks.com/v1/chain/send_transaction"
        
        elif network == "mainnet":
            self.sign_link = "https://cpu.neftyblocks.com/"
            self.push_link = "https://wax.neftyblocks.com/v1/chain/send_transaction"
        
        else:
            raise ValueError("Unknown network. Must be 'testnet' or 'mainnet'")

    def push(self) -> dict:
        signed = self.trx.get_trx_extend_info()
        signatures = signed['signatures']

        # sign with neftyblocks
        self.scraper.options(self.sign_link)
        sign_packed = self.scraper.post(self.sign_link, json={"transaction": signed['packed']}).json()

        if sign_packed.get('error'):
            raise PayWithPushError(sign_packed['error'])

        signatures += sign_packed.json()['signatures']

        # push transaction
        push = self.wax.post(self.push_link, json={
            "signatures": signatures,
            "compression": 0,
            "packed_context_free_data": "",
            "packed_trx": signed['packed']
        })
        return push

class CustomPayer:
    def __init__(self, trx, payer_client, permission="active"):
        self.trx = trx
        self.wax = self.trx.wax
        self.payer_client = payer_client


        if self.trx.actions[0].result["account"] != "abuztradewax" or\
           self.trx.actions[0].result["name"] != "noop" or\
           self.trx.actions[0].result["authorization"][0]["actor"] != payer_client.name or\
           self.trx.actions[0].result["authorization"][0]["permission"] != permission:
            
            self.trx.actions = [
                Contract("abuztradewax", actor=payer_client.name, permission=permission).noop()
            ] + self.trx.actions

    def push(self) -> dict:
        signed = self.trx.get_trx_extend_info()
        signatures = signed['signatures']

        if self.payer_client.type == 'private_key':
            chain_info = self.trx.wax.get('chain.get_info')
            
            digest = sig_digest(bytearray(signed['serealized']), chain_info['chain_id'])

            signatures.append(self.payer_client.sign(digest))
        else:
            signatures += self.payer_client.sign(bytearray(signed['serealized']))

        # push transaction
        push = self.wax.post('chain.push_transaction', json={
            "signatures": signatures,
            "compression": 0,
            "packed_context_free_data": "",
            "packed_trx": signed['packed']
        })
        return push

class PayWith:
    def __init__(self, trx, pay_with="nefty", custom_payer_client=None, network="mainnet"):
        self.client = trx.client
        self.trx = trx
        if pay_with.lower() == "nefty":
            self.pay_with = Nefty(trx, network=network)

        elif pay_with.lower() == "atomichub":
            self.pay_with = AtomicHub(trx, network=network)

        elif pay_with.lower() == "custom":
            if not custom_payer_client:
                raise ValueError("You must specify custom_payer_client")
            self.pay_with = CustomPayer(trx, custom_payer_client)

        else:
            raise ValueError("Unknown payer. Must be 'Nefty', 'AtomicHub' or 'Custom'")

    def __str__(self):
        actions = ',\n        '.join([str(x) for x in self.pay_with.trx.actions])
        return f"""litewax.Transaction(
    node={self.client.node},
    sender={self.client.name},
    actions=[
        {actions}
    ]
)"""

    def push(self):
        return self.pay_with.push()



import os
from .abigen import abigen

def Contract(name: str, client=None, actor=None, permission="active", force_recreate=False, node=None):
    if not node and client:
        node = client.node
    else:
        node = "https://wax.greymass.com"
    
    
    if not os.path.exists(f'contracts/{name.replace(".", "_")}.py') or force_recreate:
        abigen().gen(name)

    mod = __import__(f"contracts.{name.replace('.', '_')}", fromlist=[name.replace('.', '_')])
    klass = getattr(mod, name.replace('.', '_'))
    if client:
        return klass(actor=client.name, node=node, permission=permission)
    elif actor:
        return klass(actor=actor, node=node, permission=permission)

    return klass(node=node, permission=permission)

if __name__ == "__main__":
    c = Contract("res.pink")
    c.set_actor("zknmi.wam")
    print(c.noop())


class Payers:
    NEFTY = "nefty"
    NEFTYBLOCKS = "nefty"
    NeftyBlocks = "nefty"
    nefty = "nefty"
    neftyblocks = "nefty"
    Nefty = "nefty"

    ATOMICHUB = "atomichub"
    ATOMIC_HUB = "atomichub"
    AtomicHub = "atomichub"
    atomichub = "atomichub"

    CUSTOM = "custom"
    Custom = "custom"
    custom = "custom"
    




import os
import requests

file_start = """from __future__ import annotations
import eospy.cleos
import eospy.keys
import pytz
import datetime as dt
from typing import Tuple, Any

class Action:
    def __init__(self, contract: {name}, action: str, args: dict):
        self.contract = contract
        self.action = action
        self.args = args

        self.result = self()
    
    def __str__(self):
        return f"{name}::{self.action}({self.args})"
    
    def __repr__(self):
        return self.__str__()

    def __call__(self):
        return self.contract.call(self.action, self.args)

class {name}:
    def __init__(self, actor: str="", permission: str="active", node: str="https://wax.greymass.com"):
        self.wax = eospy.cleos.Cleos(url=node, version='v1')
        self.actor = actor
        self.permission = permission

    def __str__(self):
        return f"{name}(actor={self.actor}, permission={self.permission}, node={self.wax.url})"

    def set_actor(self, actor: str):
        self.actor = actor

    def generatePayload(self, account: str, name: str) -> dict:
        if self.actor == "":
            raise ValueError("actor is not set")
        return {
            "account": account,
            "name": name,
            "authorization": [{
                "actor": self.actor,
                "permission": self.permission,
            }],
        }

    def return_payload(self, payload, args) -> dict:
        # if args is empty, return payload
        if not args:
            payload['data'] = ''
            return payload
            
        data = self.wax.abi_json_to_bin(
            payload['account'], 
            payload['name'], 
            args
        )
        payload['data'] = data['binargs']
        return payload

    # ACTIONS
    def call(self, action: str, args: dict) -> dict:
        \"\"\"
        ## CALL ANY ACTION
        - Parametrs:
            - action: str
            - args: dict

        - Returns:
            - dict
        \"\"\"

        base = self.generatePayload("{normal_name}", action)

        return self.return_payload(base, args)

"""

file_action = """
    def {action}(self, {args}) -> dict:
        \"\"\"
        ## ACTION: {name}.{action}
        - Parametrs:
            {description}

        - Returns:
            - dict
        \"\"\"

        {action}_args = {genargs}
        return Action(self, "{action}", {action}_args)

"""

file_final = """    # ACTIONS END

    def push_actions(self, private_keys: Any, *actions) -> Tuple[dict, bool]:
        trx = {
            "actions": [a.result for a in list(actions)]
        }
            
        trx['expiration'] = str(
            (dt.datetime.utcnow() + dt.timedelta(seconds=60)).replace(tzinfo=pytz.UTC))

        if isinstance(private_keys, str):
            private_keys = eospy.keys.EOSKey(private_keys)

        elif isinstance(private_keys, list):
            private_keys = [eospy.keys.EOSKey(i) for i in private_keys]

        resp = self.wax.push_transaction(trx, private_keys, broadcast=True)
        return resp, True

    def create_trx(self, private_key: str, **actions) -> Tuple[dict, dict]:
        ac = []
        for k, v in actions.items():
            ac.append(eval(f'self.{k}({", ".join({v})})'))
        return self.push_actions(private_key, ac)

if __name__ == '__main__':
    contract = {name}(actor="")
    contract.push_actions(
        "PRIVATE_KEY",
        contract.transfer(
            _from="",
            to="",
            memo="",
            symbol=""
        )
    )

"""

TYPES = {
    'name': 'str',
    'asset': 'str',
    'uint64': 'int',
    'uint32': 'int',
    'uint16': 'int',
    'uint8': 'int',
    'int64': 'int',
    'int32': 'int',
    'int16': 'int',
    'int8': 'int',
    'float64': 'float',
    'float32': 'float',
    'float128': 'float',
    'bool': 'bool',
    'string': 'str',
    'public_key': 'str',
    'signature': 'str'
}

def check_ban(text):
    banwords = {
        "from": "_from",
        "for": "_for"
    }

    if text not in banwords.keys():
        return text
    else:
        return banwords.get(text)

class abigen():
    def __init__(self, node: str="https://wax.greymass.com"):
        self.node = node

    def gen(self, name):
        actions = self.get_abi(name)
        out = file_start
        for action in actions:
            if action['name'].isupper(): # it's a table
                continue

        for action in actions:
            amtext = file_action
            if action['name'].isupper(): # it's a table
                continue

            if action.get('fields', {}): 

                formargs = []
                for x in action.get('fields', {}):
                    formargs.append(f"{check_ban(x['name'])}: {TYPES.get(x['type'], 'str') if '[]' not in x['type'] else 'list'}")

                args = ', '.join(formargs)

                desc = '\n            '.join([f"- {x['name']}: {x['type']}" for x in action.get('fields', {})])

            else:
                args= ''
                desc = ''

            action_args = "{\n" + str(',\n'.join([f"            \"{x['name']}\": {check_ban(x['name'])}" for x in action.get('fields', {})])) + "\n        }"

            for text, target in [
                ['{action}', action['name']],
                ['{description}', desc],
                ['{args}', args],
                ['{name}', name],
                ['{genargs}', action_args]
            ]:
                amtext = amtext.replace(text, target)
            
            out += amtext

        out += file_final

        out = out.replace('{normal_name}', name)
        out = out.replace('{name}', name.replace('.', '_'))

        if not os.path.exists('contracts'):
            os.makedirs('contracts')

        with open('./contracts/' + name.replace('.', '_') + '.py', "w", encoding='utf-8') as f:
            f.write(out)
        return out
        
    def get_abi(self, account_name: str):
        r = requests.post(f"{self.node}/v1/chain/get_abi", json={"account_name": account_name}).json()
        return r['abi']['structs']

    def get_tx_info(self, tx: str):
        return requests.post(f"{self.node}/v1/history/get_transaction",
                   json={"id": tx, "block_num_hint": 0}).json()

if __name__ == "__main__":
    app = abigen()
    app.gen("res.pink")


from typing import List
from eospy.types import Transaction
from eospy.utils import sig_digest
import pytz
import datetime as dt

from .client import Client
from .contract import Contract
from .exceptions import *
from .paywith import PayWith


class MultiSigClient():
    """
    ### Methods:
    - SetNode
    - Transaction
    - Contract
    """
    def __init__(self, 
            private_keys: list = [], 
            cookies: list = [],  
            clients: List[Client] = [],
            node: str='https://wax.greymass.com'):

        self.node = node

        if not cookies and not private_keys and not clients:
            raise AuthNotFound("You must provide a private key, a cookie or a clients")

        self.clients = list(clients)

        for private_key in private_keys:
            self.clients.append(Client(private_key=private_key, node=node))

        for cookie in cookies:
            self.clients.append(Client(cookie=cookie, node=node))

        self.Contract = Contract

    def SetNode(self, node: str):
        self.node = node
        for client in self.clients:
            client.SetNode(node)

    def Transaction(self, *actions):
        return TX(self, *actions, node=self.node)

    def __getitem__(self, index):
        return self.clients[index]

    def __len__(self):
        return len(self.clients)
    
    def __iter__(self):
        return iter(self.clients)
    
    def __next__(self):
        return next(self.clients)

class TX():
    """
    ### Methods:
    - pay_with
    - get_trx_extend_info
    - push
    """
    def __init__(self, client: MultiSigClient, *actions, node: str='https://wax.greymass.com'):
        self.client = client
        self.node = node
        self.wax = self.client[0].wax

        self.actions = list(actions)

    def pay_with(self, payer: str, network='mainnet'):
        return PayWith(self, payer, network)

    def get_trx_extend_info(self):
        trx_wallets = []
        for action in list(self.actions):
            trx_wallets.append(action.result['authorization'][0]['actor'])

        transaction = {
            "actions": [a.result for a in self.actions]
        }
            
        transaction['expiration'] = str(
            (dt.datetime.utcnow() + dt.timedelta(seconds=60)).replace(tzinfo=pytz.UTC))

        chain_info, lib_info = self.wax.get_chain_lib_info()
        trx = Transaction(transaction, chain_info, lib_info)

        digest_anchor = sig_digest(trx.encode(), chain_info['chain_id'])

        signatures = []

        for cl in self.client:
            if cl.name not in trx_wallets:
                continue

            if cl.type == 'private_key':
                signatures.append(cl.sign(digest_anchor))

            if cl.type == 'cookie':
                signatures += cl.sign(trx.encode())

        return {
            "signatures": signatures,
            "packed": trx.encode().hex(),
            "serealized": [x for x in trx.encode()]
        }

    def push(self):
        info = self.get_trx_extend_info()
        signatures = info['signatures']
        packed = info['packed']

        push_create_offer = self.wax.post("chain.push_transaction",
            json={
                "signatures": signatures,
                "compression": 0,
                "packed_context_free_data": "",
                "packed_trx": packed
            },
            timeout=30
        )
        
        if push_create_offer['transaction_id'] == '':
            if push_create_offer['error']["what"] == 'Transaction exceeded the current CPU usage limit imposed on the transaction':
                raise CPUlimit('Error: CPU usage limit!!')

            elif push_create_offer['error']["what"] == 'Expired Transaction':
                raise ExpiredTransaction('Error: Expired Transaction!!')
            else:
                raise UnknownError(
                    f'Error: {push_create_offer["error"]["details"][0]["message"]}')

        return push_create_offer


if __name__ == "__main__":
    client = MultiSigClient(
        private_keys=["5JJYyiPpopRaQs1o7wQE6X7X1V5mc1pcE6iXGLKCQ8YpdVfv7hL"],
        #cookies=["hY0u8DUYNXjucimdWDajkuX9cLGjNlNjukaV60JZ"]
    )
    print(client[0].__dict__)
    trx = client.Transaction(
        Contract("res.pink", client[0]).noop(),

    )
    print(trx.push())


from .contract import Contract
from .abigen import abigen
from .client import Client
from .multisigclient import MultiSigClient
from .paywith import Payers

__all__ = [
    'Contract', 
    'Client',
    'MultiSigClient',
    'Payers',
    'abigen'
]

__author__ = 'abuztrade'
__version__ = '0.1.7'
__email__ = 'abuztrade.work@gmail.com'